from abc import ABC, abstractmethod

import astropy.units as u


class Encoder(ABC):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def get_reading(self) -> u.Quantity:
        ...



__all__ = ["configure", "config"]

import difflib
import os
import shutil
from collections import UserDict
from dataclasses import dataclass
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Callable, Dict, List

import astropy.units as u
from astropy.coordinates import EarthLocation
from tomlkit.toml_file import TOMLFile

from neclib import logger


@dataclass
class EnvVarName:
    necst_root: str = "NECST_ROOT"
    ros2_ws: str = "ROS2_WS"
    domain_id: str = "ROS_DOMAIN_ID"
    record_root: str = "NECST_RECORD_ROOT"


class _ConfigParsers(UserDict):
    def __missing__(self, key: str) -> Any:
        if key in self.keys():
            return self[key]
        similar_keys = difflib.get_close_matches(key.lower(), self.keys())
        if similar_keys:
            logger.info(
                f"Parser for '{key}' not found, instead using raw value. "
                f"Similarly named parameters : {similar_keys}"
            )
        return lambda x: x


class Configuration:
    """NECST configuration.

    This parameter collection supports flexible name look-up. If you query the parameter
    'antenna' by ``config.antenna``, all parameters prefixed by 'antenna' will be
    extracted.

    Examples
    --------
    >>> neclib.config.observatory
    'OMU1P85M'
    >>> neclib.config.antenna_pid_param_az
    [1.5, 0.0, 0.0]
    >>> neclib.config.antenna_pid_param
    SimpleNamespace(az=[1.5, 0.0, 0.0], el=[1.5, 0.0, 0.0])
    >>> neclib.config.antenna_pid_param.az
    [1.5, 0.0, 0.0]

    """

    _instance = None

    DefaultNECSTRoot = Path.home() / ".necst"
    DefaultConfigPath = DefaultNECSTRoot / "config.toml"

    def __new__(cls):
        if cls._instance is None:
            return super().__new__(cls)
        return cls._instance

    def __init__(self):
        self.__parameters = []
        self.reload()

    @staticmethod
    def __avoid_tomlkit_bug_printing_array(item):
        from tomlkit.items import Array

        return list(item) if isinstance(item, Array) else item

    def __repr__(self) -> str:
        length = max(len(p) for p in self.__parameters)

        def _prettify(key):
            value = self.__dict__.get(key, None)
            value = self.__avoid_tomlkit_bug_printing_array(value)
            return f"    {key:{length+2}s}{value}    ({type(value).__name__})"

        _parameters = "\n".join([_prettify(k) for k in self.__parameters])
        return f"NECST configuration\n{_parameters}"

    def __str__(self) -> str:
        def _format(key):
            value = self.__dict__[key]
            value = self.__avoid_tomlkit_bug_printing_array(value)
            return f"{key}={value}"

        _parameters = ", ".join([_format(k) for k in self.__parameters])
        return f"Configuration({_parameters})"

    def reload(self):
        """Reload the parameters, in case the config file is updated.

        .. WARNING::

            Manually set parameters will be lost on reload.

        """
        for param in self.__parameters:
            del self.__dict__[param]
        self.__config_path = self.__find_config_path()
        self.__dotnecst = self.__config_path.parent
        self.__parameters = self.__parse()

    def __getattr__(self, key: str) -> Any:
        prefix = key + "_"
        prefix_length = len(prefix)
        match = {
            k[prefix_length:]: getattr(self, k)
            for k in self.__parameters
            if k.startswith(prefix)
        }
        return SimpleNamespace(**match) if match else None

    def __get_parser(self) -> _ConfigParsers:
        _parsers: Dict[str, Callable[[Any], Any]] = {
            "observatory": str,
            "location": lambda x: EarthLocation(**x),
            "simulator": bool,
            "antenna_pid_param_az": list,
            "antenna_pid_param_el": list,
            "antenna_drive_range_az": lambda x: list(map(u.Quantity, x)),
            "antenna_drive_range_el": lambda x: list(map(u.Quantity, x)),
            "antenna_drive_warning_limit_az": lambda x: list(map(u.Quantity, x)),
            "antenna_drive_warning_limit_el": lambda x: list(map(u.Quantity, x)),
            "antenna_drive_critical_limit_az": lambda x: list(map(u.Quantity, x)),
            "antenna_drive_critical_limit_el": lambda x: list(map(u.Quantity, x)),
            "antenna_pointing_accuracy": u.Quantity,
            "antenna_pointing_parameter_path": lambda x: self.__dotnecst / Path(x),
            "antenna_max_acceleration_az": u.Quantity,
            "antenna_max_acceleration_el": u.Quantity,
            "antenna_max_speed_az": u.Quantity,
            "antenna_max_speed_el": u.Quantity,
            "antenna_speed_to_pulse_factor_az": u.Quantity,
            "antenna_speed_to_pulse_factor_el": u.Quantity,
            "antenna_command_frequency": int,
            "antenna_command_offset_sec": float,
            "antenna_encoder_port_az": str,
            "antenna_encoder_port_el": str,
            "ros_service_timeout_sec": float,
            "ros_communication_deadline_sec": float,
        }
        return _ConfigParsers(_parsers)

    def __parse(self) -> List[str]:
        raw_config = TOMLFile(self.__config_path).read()
        parser = self.__get_parser()
        for k, v in raw_config.items():
            setattr(self, k, parser[k](v))
        return list(raw_config.keys())

    def __find_config_path(self) -> Path:
        candidates = [self.DefaultNECSTRoot]
        if EnvVarName.necst_root in os.environ.keys():
            candidates.insert(0, Path(os.environ[EnvVarName.necst_root]))

        for path in candidates:
            config_path = path / "config.toml"
            if config_path.exists():
                logger.info(f"Imported configuration file '{config_path}'")
                return config_path
        logger.error(
            "Config file not found, using the default parameters. "
            "To create the file with default parameters, run `neclib.configure()`."
        )
        return Path(__file__).parent / "src" / "config.toml"

    @classmethod
    def configure(cls) -> None:
        """Create config file under ``$HOME/.necst``"""
        cls.DefaultNECSTRoot.mkdir(exist_ok=True)
        for filename in ["config.toml", "pointing_param.toml"]:
            _target_path = cls.DefaultNECSTRoot / filename
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(Path(__file__).parent / "src" / filename, _target_path)
        cls().reload()


configure = Configuration.configure
config = Configuration()



__all__ = ["DBWriter"]

import queue
import time
from datetime import datetime
from pathlib import Path
from threading import Event, Thread
from typing import Any, Dict, List, Optional, Sequence, Tuple

import necstdb

from ..typing import PathLike


class DBWriter:
    def __init__(self, record_root: PathLike):
        self.record_root = Path(record_root)
        self.db_name: Optional[str] = None

        self.db: Optional[necstdb.necstdb.necstdb] = None
        self.tables: Dict[str, necstdb.necstdb.table] = {}
        self.table_last_update: Dict[str, float] = {}

        self.data_queue = queue.Queue()
        self.thread = None

    def start_recording(self, name: str = None):
        self.db_name = name
        self.db_dir = self.record_root / (name or self._get_date_str())

        self.db = necstdb.opendb(self.db_dir, mode="w")
        self._stop_event = Event()
        self.thread = Thread(target=self._update_background, daemon=True)
        self.thread.start()

    def append(self, data: Dict[str, Any]) -> None:
        self.data_queue.put(data)

    def stop_recording(self):
        _ = [table.close() for table in self.tables.values()]
        self.tables.clear()
        self.table_last_update.clear()

        self._stop_event.set()
        self.thread = None

    def _update_background(self):
        while not self._stop_event.is_set():
            self._check_date()
            self._check_liveliness()
            if self.data_queue.empty():
                time.sleep(0.01)
                continue
            data = self.data_queue.get()
            self._write(data["topic"], data["slots"])

    def _write(self, name: str, slots: Dict[str, Any]):
        name = self._validate_name(name)
        now = time.time()
        self.table_last_update[name] = now

        table_data = [now]
        table_info = [{"key": "received_time", "format": "d", "size": 8}]
        for slot in slots:
            data, info = self._parse_slot(slot)
            table_data.extend(data)
            table_info.append(info)
        if name not in self.tables.keys():
            self.add_table(
                name,
                {"data": table_info, "memo": f"Generated by {self.__class__.__name__}"},
            )
        self.tables[name].append(*table_data)

    def _parse_slot(self, slot: Dict[str, Any]) -> Tuple[List[Any], Dict[str, str]]:
        def str_to_bytes(data: str) -> bytes:
            try:
                return data if isinstance(data, bytes) else data.encode("utf-8")
            except AttributeError:
                return (str_to_bytes(elem) for elem in data)

        conversion_table = {
            "bool": lambda dat: (dat, "?", 1),
            "byte": lambda dat: (dat, f"{len(dat)}s", 1 * len(dat)),
            "char": lambda dat: (dat, "c", 1),
            "float32": lambda dat: (dat, "f", 4),
            "float64": lambda dat: (dat, "d", 8),
            "int8": lambda dat: (dat, "b", 1),
            "int16": lambda dat: (dat, "h", 2),
            "int32": lambda dat: (dat, "i", 4),
            "int64": lambda dat: (dat, "q", 8),
            "uint8": lambda dat: (dat, "B", 1),
            "uint16": lambda dat: (dat, "H", 2),
            "uint32": lambda dat: (dat, "I", 4),
            "uint64": lambda dat: (dat, "Q", 8),
            "string": lambda dat: (str_to_bytes(dat), f"{len(dat)}s"),
        }
        for k in conversion_table.keys():
            if slot["type"].find(k) != -1:
                data, format_, size = conversion_table[k](slot["value"])
        if isinstance(data, Sequence) and (not isinstance(data, (str, bytes))):
            length = len(data)
            if format_.find("s") == 0:
                format_ = "0s" if length == 0 else format_ * length
                # Because 5s5s5s is ok, but 35s has different meaning.
            else:
                format_ = f"{length}{format_}"
            size *= length
            data = list(data)
        else:
            data = [data]
        return data, {"key": slot["key"], "format": format_, "size": size}

    def _validate_name(self, name: str) -> str:
        return name.replace("/", "-").strip("-")

    def _get_date_str(self) -> str:
        return datetime.utcnow().strftime("%Y%m/%Y%m%d.necstdb")

    def _check_date(self) -> None:
        if (self.db_name is None) and (self.db_dir.name not in self._get_date_str()):
            self.stop_recording()
            self.start_recording()

    def _check_liveliness(self) -> None:
        table_names = self.tables.keys()
        now = time.time()
        for name in table_names:
            if now - self.table_last_update[name] > 60:
                self.remove_table(name)

    def add_table(self, name: str, header: Dict[str, Any]) -> None:
        name = self._validate_name(name)
        if name in self.tables:
            raise ValueError(f"Table '{name}' already opened.")
        if name not in self.db.list_tables():
            header.update({"necstdb_version": necstdb.__version__})
            self.db.create_table(name, header)
        self.tables[name] = self.db.open_table(name, mode="ab")
        self.table_last_update[name] = time.time()  # Temporal value.

    def remove_table(self, name: str) -> None:
        name = self._validate_name(name)
        if name in self.tables:
            self.tables[name].close()
        _ = self.tables.pop(name, None)
        _ = self.table_last_update.pop(name, None)



__all__ = ["DriveLimitChecker"]

from typing import List, Tuple

import astropy.units as u

from neclib import logger
from ..typing import QuantityValue, Unit
from .. import utils


class DriveLimitChecker:
    """Find optimum and safe angle to drive to.

    The optimization includes:

    1. Apply drive range constraint
    2. Avoid direction unwrapping (360deg drive) during observation
    3. Over-180deg drive

    Parameters
    ----------
    limit
        Drive range limit, the optimization result never go out of this range.
    preferred_limit
        Drive range limit, which can be violated to continue observation or no other
        choice found.
    unit
        Angular unit in which the parameters are given.
    max_observation_size
        If separation between current and target coordinates is smaller than this value,
        direction unwrapping won't occur, possibly violating ``preferred_limit``, to
        not interrupt observation.

    Examples
    --------
    >>> checker = neclib.coordinates.DriveLimitChecker(
    ...     [-260 << u.deg, 260 << u.deg], [-250 << u.deg, 250 << u.deg]
    ... )
    >>> checker.optimize(current=-200 << u.deg, target=170 << u.deg)
    <Quantity -190. deg>

    """

    def __init__(
        self,
        limit: Tuple[QuantityValue, QuantityValue],
        preferred_limit: Tuple[QuantityValue, QuantityValue] = None,
        *,
        unit: Unit = None,
        max_observation_size: QuantityValue = 5 << u.deg,
    ) -> None:
        self.limit = utils.get_quantity(*limit, unit=unit)
        self.max_observation_size = utils.get_quantity(max_observation_size, unit=unit)
        if preferred_limit is None:
            self.preferred_limit = self.limit
        else:
            self.preferred_limit = utils.get_quantity(*preferred_limit, unit=unit)

        _limit_ascending = self.limit[0] < self.limit[1]
        _preferred_limit_ascending = self.preferred_limit[0] < self.preferred_limit[1]
        if (not _limit_ascending) or (not _preferred_limit_ascending):
            # Silently swapping the values could cause destructive result, i.e.
            # [270deg, 90deg] can mean [-90deg, 90deg], which is opposite to
            # [90deg, 270deg].
            raise ValueError("Limits should be given in ascending order.")

    def _get_candidates(self, target: u.Quantity) -> List[u.Quantity]:
        # NOTE: Author has made no effort making this algorithm readable, but the logic
        # isn't so complicated.
        _360deg = 360 << u.deg
        min_target_candidate = target - _360deg * ((target - self.limit[0]) // _360deg)
        target_candidates = [
            angle
            for angle in utils.frange(min_target_candidate, self.limit[1], _360deg)
        ]
        return target_candidates

    def _select(self, current: u.Quantity, candidates: List[u.Quantity]) -> u.Quantity:
        if len(candidates) < 1:
            return None
        if len(candidates) == 1:
            return candidates[0]

        ordered_by_distance = sorted(
            [(t, abs(t - current)) for t in candidates], key=lambda x: x[1]
        )
        for target, distance in ordered_by_distance:
            if distance < self.max_observation_size:
                return target
            if self.preferred_limit[0] < target < self.preferred_limit[1]:
                return target

    def optimize(
        self,
        current: QuantityValue,
        target: QuantityValue,
        unit: Unit = None,
    ) -> u.Quantity:
        """Optimize the coordinate to command.

        Parameters
        ----------
        current
            Current coordinate.
        target
            Target coordinate, to be optimized.
        unit
            Angular unit in which the ``current`` and ``target`` are given.

        """
        current, target = utils.get_quantity(current, target, unit=unit)

        target_candidates = self._get_candidates(target)
        optimized = self._select(current, target_candidates)
        self._warn_unpreferred_result(optimized)
        return optimized

    def _warn_unpreferred_result(self, result: u.Quantity) -> None:
        lower = self.limit[0] < result < self.preferred_limit[0]
        upper = self.limit[1] > result > self.preferred_limit[1]
        if lower or upper:
            logger.warning("Command position is near drive range limit.")



"""Type aliases for simple type hinting."""

__all__ = ["PathLike", "AngleUnit"]

import os
from typing import Literal, Union

import astropy.units as u
import numpy as np


PathLike = Union[os.PathLike, str]
"""Alias of ``os.PathLike``, with ``str`` type combined."""

AngleUnit = Literal["deg", "arcmin", "arcsec", "rad"]
"""Literal expression of supported angular units in this package."""

Number = Union[int, float, np.ndarray]
"""Number types."""

Boolean = Union[bool, np.bool_]
"""Boolean types."""

Unit = Union[str, u.Unit]
"""Unit of physical quantity."""

QuantityValue = Union[Number, u.Quantity]
"""Physical quantity or primitive type with unit separately specified."""



__all__ = ["ND287"]

import astropy.units as u
import ogameasure

from .encoder_base import Encoder


class ND287(Encoder):
    """Encoder readout.

    Parameters
    ----------
    port
        Device path in UNIX file system, like ``"/dev/ttyUSB1"``.

    """

    Manufacturer = "HEIDENHAIN"
    Model = "ND287"

    def __init__(self, port: str) -> None:
        self.driver = ogameasure.HEIDENHAIN.ND287(port)

    def get_reading(self) -> u.Quantity:
        raw = self.driver.output_position_display_value()
        return float(raw.strip(b"\x02\x00\r\n").decode())
